#include "VisibilityManager.h"
#include "StopAndWaitCullingManager.h"
#include "CoherentHierarchicalCullingManager.h"
#include "FrustumCullingManager.h"
#include "DummyPreprocessingManager.h"
#include "DummyQueryManager.h"

namespace GtpVisibility {

//-----------------------------------------------------------------------
VisibilityManager::VisibilityManager(VisibilityEnvironment *visEnvironment):
mVisibilityThreshold(0), mCullingManager(NULL),
mQueryManager(new DummyQueryManager(NULL)),
mPreprocessingManager(new DummyPreprocessingManager(NULL)),
mVisibilityEnvironment(visEnvironment)
{
	SetCullingManager(VisibilityEnvironment::COHERENT_HIERARCHICAL_CULLING);
	//SetCullingManager(VisibilityEnvironment::STOP_AND_WAIT_CULLING);
	//SetCullingManager(VisibilityEnvironment::FRUSTUM_CULLING);
}
//-----------------------------------------------------------------------
VisibilityManager::~VisibilityManager()
{
	delete mCullingManager;
}
//-----------------------------------------------------------------------
void VisibilityManager::SetCullingManager(
	VisibilityEnvironment::CullingManagerType ocmType)
{
	// change only if necessary
	if (!mCullingManager || (mCullingManagerType != ocmType))
	{
		// delete old culling manager
		if (mCullingManager)
		{
			delete mCullingManager;
		}
		mCullingManagerType = ocmType;
		
		switch (ocmType)
		{
		case VisibilityEnvironment::STOP_AND_WAIT_CULLING:
			mCullingManager = new StopAndWaitCullingManager();
			break;

		case VisibilityEnvironment::COHERENT_HIERARCHICAL_CULLING:
			mCullingManager = new CoherentHierarchicalCullingManager();
			break;

		case VisibilityEnvironment::FRUSTUM_CULLING:
			mCullingManager = new FrustumCullingManager();
			break;

		default:
			mCullingManager = new StopAndWaitCullingManager();
			break;

		}

		// -- set threshold
		mCullingManager->SetVisibilityThreshold(mVisibilityThreshold);		
	}
}
//-----------------------------------------------------------------------
void VisibilityManager::SetVisibilityCullingThreshold(unsigned int visibilityThreshold)
{
	mVisibilityThreshold = visibilityThreshold;
	mCullingManager->SetVisibilityThreshold(mVisibilityThreshold);
}
//-----------------------------------------------------------------------
CullingManager *VisibilityManager::GetCullingManager()
{
	return mCullingManager;
}
//-----------------------------------------------------------------------
void VisibilityManager::ApplyVisibilityCulling()
{
	mCullingManager->RenderScene();
}
//-----------------------------------------------------------------------
VisibilityEnvironment::CullingManagerType VisibilityManager::GetCullingManagerType()
{
	return mCullingManagerType;
}

} // namespace GtpVisibility
